﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.IO;
using System.Diagnostics;
using System.Reflection;
using junit;

namespace c4ju
{
    class Program
    {
        static int Main(string[] args)
        {
            bool input_stdin = false;
            bool show_worst = false;
            bool verbose = false;
            List<string> files = new List<string>();
            string outfile = "test_default.xml";
            string filter_xmlpath = "filter_default.xml";
            List<string> excluding = new List<string>();
            string excluding_list_txt = null;

            #region コマンドライン解析
            // コマンドライン引数の解析
            for(int i=0; i < args.Length; ++i )
            {
                string argv = args[i];
                if (argv.IndexOf("--") == 0)
                {
                    string option = argv.Substring(2);
                    string opt_out_xmlfile = "out_xmlfile";
                    string opt_filter = "filter";
                    string opt_excluding = "excluding";
                    #region 出力ファイル
                    if (option.IndexOf(opt_out_xmlfile) == 0)
                    {
                        // 出力ファイル名の指定
                        if (option.Length > 1)
                        {
                            if (option.IndexOf("=") == opt_out_xmlfile.Length)
                                outfile = option.Substring(opt_out_xmlfile.Length + 1);
                        }
                        else if( i < args.Length )
                        {
                            outfile = args[++i];
                        }
                        if( outfile[0] == '\"' )
                        {
                            outfile = outfile.Substring(1, outfile.Length - 2);
                        }
                    }
                    #endregion
                    #region ヘルプ
                    else if (option == "help")
                    {
                        Console.WriteLine(
@"c4ju [options] file1.xml ...
Process files listed on command line.
If the filenames include '-', read a list of files from standard input.

Command Line Options:
--help                  : generate this help message.
--out_xmlfile=<path>    : path of xml report.
--filter=<path>         : path of test filter xml.
--excluding=<file>      : read the ignore list from a file.
--verbose               : displays detailed messages.
"
                        );
                        return 0;
                    }
                    #endregion
                    #region フィルターファイル
                    else if (option.IndexOf(opt_filter) == 0)
                    {
                        // フィルターファイル名の指定
                        if (option.Length > 1)
                        {
                            if (option.IndexOf("=") == opt_filter.Length)
                                filter_xmlpath = option.Substring(opt_filter.Length + 1);
                        }
                        else if (i < args.Length)
                        {
                            filter_xmlpath = args[++i];
                        }

                        if (filter_xmlpath[0] == '\"')
                        {
                            filter_xmlpath = filter_xmlpath.Substring(1, filter_xmlpath.Length - 2);
                        }
                    }
                    #endregion
                    #region ワースト情報の出力
                    else if (option == "show_worst")
                    {
                        show_worst = true;
                    }
                    #endregion
                    #region 無視リスト
                    else if (option.IndexOf(opt_excluding) == 0)
                    {
                        // フィルターファイル名の指定
                        if (option.Length > 1)
                        {
                            if (option.IndexOf("=") == opt_excluding.Length)
                                excluding_list_txt = option.Substring(opt_excluding.Length + 1);
                        }
                        else if (i < args.Length)
                        {
                            filter_xmlpath = args[++i];
                        }

                        if (excluding_list_txt[0] == '\"')
                        {
                            excluding_list_txt = excluding_list_txt.Substring(1, excluding_list_txt.Length - 2);
                        }
                    }
                    #endregion
                    #region verbose
                    else if (option == "verbose")
                    {
                        verbose = true;
                    }
                    #endregion
                }
                else if( argv[0] == '-' )
                {
                    #region バージョン
                    if (argv == "-v")
                    {
                        AssemblyName asmName = Assembly.GetEntryAssembly().GetName();
                        Version ver = asmName.Version;
                        Console.Write(asmName.Name);
                        Console.Write(" v"); 
                        Console.WriteLine(ver);
                        return 0;
                    }
                    #endregion
                    #region 標準入力ファイル
                    else if (i == args.Length - 1 && argv.Length == 1)
                    {
                        input_stdin = true;
                    }
                    #endregion
                }
                #region 入力ファイル
                else
                {
                    files.Add(argv);
                }
                #endregion
            }
            #endregion

            #region フィルター用意
            List<Filter> filter_list = FilterBuilder.Build(filter_xmlpath);
            if (filter_list == null) return 1;
            #endregion

            #region 標準入力からのファイルパス取得
            if (input_stdin)
            {
                string line = null;
                while ( Console.In.Peek() != 0 && (line = Console.In.ReadLine()) != null)
                {
                    files.Add(line);
                }
            }
            #endregion

            #region 無視リスト
            if (excluding_list_txt != null)
            {
                using (StreamReader sr = new StreamReader(excluding_list_txt))
                {
                    string line = null;
                    while ((line = sr.ReadLine()) != null)
                    {
                        excluding.Add(line);
                    }
                }
            }
            #endregion

            #region テストの開始処理
            Stopwatch test_suites_sw = new Stopwatch();
            Stopwatch test_case_sw = new Stopwatch();
            Stopwatch test_sw = new Stopwatch();

            JunitXmlDocument junitDoc = new JunitXmlDocument();
            XmlNode TestSuites = junitDoc.AppendTestSuites("AllTest"
                , 0 // 仮
                , 0 // 仮
                , 0
                , test_suites_sw.Elapsed    // 仮
                );
            int test_total_count = 0;
            int failure_total_count = 0;

            test_suites_sw.Start();
            #endregion

            #region xml の解析
            foreach (string f in files)
            {
                XmlDocument xml = new XmlDocument();
                try
                {
                    xml.Load(f);

                    if (verbose)
                    {
                        Console.WriteLine(f);
                    }

                    foreach (Filter filter in filter_list)
                    {
                        #region テストケース
                        try
                        {
                            string test_case_name = filter.TestName + Path.GetFileNameWithoutExtension(f);
                            XmlNodeList selectNodes = xml.SelectNodes(filter.strXPath);
                            if (selectNodes != null && selectNodes.Count > 0)
                            {
                                XmlNode TestSuite = junitDoc.AppendTestSuite(test_case_name
                                    , selectNodes.Count
                                    , 0 // 仮
                                    , 0
                                    , test_case_sw.Elapsed    // 仮
                                    );
                                int failure_count = 0;

                                test_case_sw.Reset();
                                test_case_sw.Start();
                                foreach (XmlNode node in selectNodes)
                                {
                                    #region テスト
                                    List<JunitXmlDocument.FailureResult> fr_list = new List<JunitXmlDocument.FailureResult>();
                                    XmlNode sourceNode = null;
                                    string name = node.InnerText;
                                    string file = "";
                                    int line = -1;
                                    #region テスト名・ファイルパス情報の取得
                                    XmlNode parentNode = node.ParentNode;
                                    if (parentNode.Name == "member_function")
                                    {
                                        name = parentNode.SelectSingleNode("name").InnerText;
                                        sourceNode = parentNode.SelectSingleNode("extent/source_reference");
                                    }
                                    else if (parentNode.Name == "module_summary")
                                    {
                                        name = "module_summary";
                                        sourceNode = parentNode.ParentNode.SelectSingleNode("module_detail/source_reference");
                                    }
                                    if (sourceNode != null)
                                    {
                                        file = sourceNode.Attributes["file"].Value;
                                        line = int.Parse(sourceNode.Attributes["line"].Value);
                                    }
                                    #endregion

                                    test_sw.Reset();
                                    test_sw.Start();
                                    foreach (FilterAttribute filter_attr in filter.list)
                                    {
                                        #region 属性ごとにテスト
                                        try
                                        {
                                            XmlAttribute attr = node.Attributes[filter_attr.Name];
                                            if (attr != null)
                                            {
                                                // フィルタリング用のxmlは式が記述されている
                                                // value は比較対象の属性valueに置換される。
                                                // 式評価がfalseを返した場合、テスト失敗とみなす。

                                                // parse 失敗はエラーとしない
                                                float lhs = float.Parse(attr.Value);
                                                try
                                                {
                                                    string result = filter_attr.Eval.Eval(lhs);
                                                    ++test_total_count;

                                                    if (!bool.Parse(result))
                                                    {
                                                        string expr = filter_attr.Value;
                                                        expr = expr.Replace("value", attr.Value);
                                                        JunitXmlDocument.FailureResult fr = new JunitXmlDocument.FailureResult();
                                                        fr.Message = "Expected : " + filter_attr.Value + "\n  actual : " + expr;
                                                        fr.File = file;
                                                        fr.Line = line;
                                                        fr_list.Add(fr);
                                                        string test_combine_name = test_case_name + "." + name;
                                                        if (verbose)
                                                        {
                                                            Console.WriteLine(test_combine_name);
                                                            Console.WriteLine(fr.Message);
                                                        }

                                                        // ワースト情報を保存
                                                        if (show_worst)
                                                        {
                                                            if (filter_attr.Worst == null)
                                                            {
                                                                filter_attr.Worst = new WorstInfo();
                                                            }
                                                            if (lhs > filter_attr.Worst.Value)
                                                            {
                                                                filter_attr.Worst.Name = test_combine_name;
                                                                filter_attr.Worst.Value = lhs;
                                                                filter_attr.Worst.file = file;
                                                                filter_attr.Worst.line = line;
                                                            }
                                                        }
                                                    }
                                                }
                                                catch (Exception e)
                                                {
                                                    Console.Error.WriteLine(e.Message);
                                                }
                                            }
                                        }
                                        catch
                                        {
                                        }
                                        #endregion
                                    }

                                    test_sw.Stop();
                                    bool except = false;
                                    if (fr_list.Count > 0)
                                    {
                                        string full_name = test_case_name + "." + name;
                                        foreach (string pattern in excluding)
                                        {
                                            if (System.Text.RegularExpressions.Regex.IsMatch(full_name, pattern))
                                            {
                                                // マッチしたら無視
                                                except = true;
                                                break;
                                            }
                                        }
                                    }
                                    XmlNode testcase = junitDoc.AppendTestCase(TestSuite
                                        , name
                                        , true
                                        , test_sw.Elapsed
                                        , fr_list
                                        );
                                    if (except)
                                    {
                                        // skip ノード追加
                                        testcase.AppendChild(junitDoc.CreateElement("skipped"));
                                    }
                                    else
                                    {
                                        if (fr_list.Count > 0)
                                        {
                                            ++failure_count;
                                        }
                                    }
                                    #endregion
                                }
                                test_case_sw.Stop();

                                failure_total_count += failure_count;
                                TestSuite.Attributes["failures"].Value = failure_count.ToString();
                                TestSuite.Attributes["time"].Value = test_case_sw.Elapsed.TotalSeconds.ToString();
                            }
                        }
                        catch (Exception e)
                        {
                            Console.Error.WriteLine(e.Message);
                        }
                        #endregion
                    }
                }
                catch (Exception e)
                {
                    Console.Error.WriteLine(e.Message);
                }
            }
            #endregion

            #region テストの終了処理
            test_suites_sw.Stop();

            TestSuites.Attributes["tests"].Value = test_total_count.ToString();
            TestSuites.Attributes["failures"].Value = failure_total_count.ToString();
            TestSuites.Attributes["time"].Value = test_suites_sw.Elapsed.TotalSeconds.ToString();

            junitDoc.Save(outfile);
            #endregion

            #region ワースト情報の出力
            if (show_worst)
            {
                Console.WriteLine("");
                Console.WriteLine("********************");
                Console.WriteLine("The largest values.");
                foreach (Filter filter in filter_list)
                {
                    Console.WriteLine(filter.TestName);
                    foreach (FilterAttribute attr in filter.list)
                    {
                        if (attr.Worst != null)
                        {
                            Console.WriteLine("    " + attr.Worst.Name);
                            if (attr.Worst.file != null)
                            {
                                Console.WriteLine("    " + attr.Worst.file + ":" + attr.Worst.line);
                            }
                            Console.WriteLine("        " + attr.Name + " : " + attr.Worst.Value.ToString());
                            Console.WriteLine("");
                        }
                    }
                }
            }
            #endregion

            return 0;
        }
    }
}

